from luqum.tree import (
    NONE_ITEM,
    UnknownOperation,
    AndOperation,
    NoneItem,
    OrOperation,
    Word,
    Phrase,
    Group,
    Prohibit,
)
from luqum.visitor import TreeTransformer


class FilterQueryTransformer(TreeTransformer):
    allowed = [
        UnknownOperation,
        AndOperation,
        OrOperation,
        Group,
        Prohibit,
        Phrase,
        Word,
        NoneItem,
    ]

    def visit_prohibit(self, node, context):
        """
        If a word is in a Prohibit (e.g. -word), we want to keep it,
        because it indicates that the word should not be
        present in the search results.
        """

        context["retain_words"] = True

        return self.generic_visit(node, context)

    def visit_word(self, node, context):
        """
        We remove all words that are not part of a phrase,
        except for those in a Prohibit (e.g. -word),
        because we're trying to limit the search to phrases and
        negations only.
        """

        if context.get("retain_words"):
            # If we're in a prohibit, we keep the word as it indicates
            # that the word should not be present in the search results.
            yield node
        else:
            yield NONE_ITEM

    def visit_or_operation(self, node, context):
        """
        If any child is a NoneItem, return NoneItem, e.g.
        OrOperation(NoneItem(), Phrase("test")) should return NoneItem,
        because NoneItem is effectively a wildcard that could match anything
        and an OR of that with anything else should still match anything.
        This removes OR in the case where one of the terms has been removed.
        """

        context["retain_words"] = True

        nodes = self.generic_visit(node, context)

        for n in nodes:
            if any(isinstance(child, NoneItem) for child in n.children):
                yield NONE_ITEM
            else:
                yield n

    def visit_and_operation(self, node, context):
        """
        If any child is a NoneItem, return the other children, e.g.
        AndOperation(NoneItem(), Phrase("test")) should return just Phrase("test").
        This removes AND in the case where one of the terms has been removed.
        """

        nodes = self.generic_visit(node, context)

        for n in nodes:
            if any(isinstance(child, NoneItem) for child in n.children):
                for child in n.children:
                    if not isinstance(child, NoneItem):
                        yield child
            else:
                yield n

    def generic_visit(self, node, context):
        nodes = super().generic_visit(node, context)

        for n in nodes:
            # Return NoneItem if all the children of the nodes are NoneItem
            if len(n.children) > 0 and all(
                isinstance(child, NoneItem) for child in n.children
            ):
                yield NONE_ITEM
            # If the node is not one of the allowed types,
            elif not isinstance(n, tuple(self.allowed)):
                # if the node has children,
                if len(n.children) > 0:
                    # yield them,
                    for child in n.children:
                        yield child
                else:
                    yield NONE_ITEM
            else:
                yield n

class WithoutNegationsTransformer(TreeTransformer):
    def visit_prohibit(self, node, context):
        yield NONE_ITEM



from django.conf import settings
from django.contrib.postgres.search import SearchQuery
from django.db import transaction
from django.db.models import F, Q, Window
from django.db.models.functions import DenseRank
from django.dispatch import receiver
from pgvector.django import CosineDistance

from torque import models as torque_models
from torque.signals import search_filter, search_index_rebuilt, update_cache_document

from semantic_search.llm import llm, local_llm
from semantic_search.models import SemanticSearchCacheDocument
from semantic_search.utils import build_semantic_summary, filter_queries

from semantic_search.embedding_cache import embeddings_cache

BATCH_SIZE = 32 * 4


@receiver(update_cache_document)
def update_semantic_cache_document(sender, **kwargs):
    cache_document = kwargs["cache_document"]
    filtered_data = kwargs["filtered_data"]
    document_dict = kwargs["document_dict"]

    with transaction.atomic():
        SemanticSearchCacheDocument.objects.filter(
            search_cache_document=cache_document
        ).delete()

        semantic_summary = build_semantic_summary(document_dict, filtered_data)

        embeddings = embeddings_cache.get(semantic_summary)

        if not embeddings:
            embeddings = local_llm.get_embeddings(semantic_summary)
            embeddings_cache.add(semantic_summary, embeddings)

        semantic_search_cache_documents = [
            SemanticSearchCacheDocument(
                search_cache_document=cache_document,
                data=semantic_summary,
                data_embedding=embedding,
            )
            for embedding in embeddings
        ]

        SemanticSearchCacheDocument.objects.bulk_create(semantic_search_cache_documents)


@receiver(search_index_rebuilt)
def rebuild_semantic_search_index(sender, **kwargs):
    wiki_config = kwargs["wiki_config"]

    search_cache_documents = torque_models.SearchCacheDocument.objects.filter(
        wiki_config=wiki_config
    )
    scds_to_fetch = []
    semantic_summaries_to_fetch = []
    llm_embeddings = []
    semantic_sc_documents = []
    for scd in search_cache_documents:
        document_dict = scd.document.to_dict(wiki_config, "latest")["fields"]
        semantic_summary = build_semantic_summary(document_dict, scd.filtered_data)

        embeddings = embeddings_cache.get(semantic_summary)

        if embeddings:
            semantic_sc_documents.append(
                SemanticSearchCacheDocument(
                    search_cache_document=scd,
                    data_embedding=embeddings,
                    data=semantic_summary,
                )
            )
        else:
            scds_to_fetch.append(scd)
            semantic_summaries_to_fetch.append(semantic_summary)

            if len(semantic_summaries_to_fetch) % BATCH_SIZE == 0:
                llm_embeddings.extend(llm.get_embeddings(semantic_summaries_to_fetch[-BATCH_SIZE:]))

    if semantic_summaries_to_fetch:
        llm_embeddings.extend(
            llm.get_embeddings(
                semantic_summaries_to_fetch[-(len(semantic_summaries_to_fetch) % BATCH_SIZE) :]
            )
        )

    for scd, (semantic_summary, embedding) in zip(scds_to_fetch,
            zip(semantic_summaries_to_fetch, llm_embeddings)
        ):
        embeddings_cache.add(semantic_summary, embedding)
        semantic_sc_documents.append(
            SemanticSearchCacheDocument(
                search_cache_document=scd,
                data_embedding=embedding,
                data=semantic_summary,
            )
        )

    SemanticSearchCacheDocument.objects.bulk_create(semantic_sc_documents)


@receiver(search_filter)
def semantic_filter(sender, **kwargs):
    similarity = getattr(settings, "SEMANTIC_SEARCH_SIMILARITY", 0.7)

    cache_documents = kwargs["cache_documents"]
    qs = kwargs.get("qs")
    qs_without_negations = filter_queries(qs, keep_negations=False)

    if qs_without_negations:
        embeddings = local_llm.get_embeddings(qs_without_negations, prompt_name="query")

        distances = {}
        semantic_qs = Q()
        for i, embedding in enumerate(embeddings):
            distance_col_name = f"distance_{i}"
            distances[distance_col_name] = CosineDistance(
                "semantic_documents__data_embedding", embedding
            )
            semantic_qs |= Q(**{f"{distance_col_name}__lte": similarity})

        # filters phrases and negations from the queries,
        # what we're calling exact or advanced search
        filter_qs = Q()
        for q in filter_queries(qs):
            if q != "":
                filter_qs &= Q(data_vector=SearchQuery(q, search_type="websearch"))

        results = (
            cache_documents
                .alias(**distances)
                .annotate(
                    score=1 - F("distance_0")
                )
                .filter(semantic_qs)
                .filter(filter_qs)
        )

        return results



import pytest
from semantic_search.utils import filter_queries


@pytest.mark.parametrize(
    "query,expected_result",
    [
        # should remove all words
        ("water in india", ""),
        # should keep negated words
        ("-test", "-test"),
        # should remove all words except negated
        ("water in india -test", "-test"),
        # should remove words, keep negated phrase
        (
            'water in india -"test 123"',
            '-"test 123"',
        ),
        # should not remove words when OR-ed
        ('water OR "test"', 'water OR "test"'),
        # should keeps a phrase
        ('water AND "test"', '"test"'),
        # should remove AND, group where none of the terms are retained
        (
            "(water AND test)",
            "",
        ),
        # should keep negated term in AND, retain group
        ("(water AND -test)", "( -test)"),
        # should remove NOT (search doesn't yet support NOT), non-negated word
        ("NOT water", ""),
        # should keep phrase, remove NOT
        (
            'NOT "test"',
            '"test"',
        ),
        # all together now
        (
            'title:water water in india -"is a resource" -goats "test"',
            '-"is a resource" -goats "test"',
        ),
    ],
)
def test_filter_queries(query, expected_result):
    results = filter_queries([query])
    assert results[0] == expected_result, (
        f"Expected {expected_result}, but got {results}"
    )


@pytest.mark.parametrize(
    "query,expected_result",
    [
        ("-test", ""),
        (
            'title:water water in india -"is a resource" -goats "test"',
            'title:water water in india "test"',
        ),
    ],
)
def test_filter_queries_without_negations(query, expected_result):
    results = filter_queries([query], keep_negations=False)
    assert results[0] == expected_result, (
        f"Expected {expected_result}, but got {results}"
    )



from django.conf import settings
import hashlib
import datetime

class EmbeddingsCache:
    """Simple LRU Cache for embeddings calls to either the local or remote LLM.

    Just in memory, so that re-runs of rebuilding the search caches are much snappier.
    For the most part, when rebuilding search caches, the document according to the
    embeddings are equivalent, and so there's no need to ask the LLM to regenerate them.

    CACHE_SIZE is in number of items, not memory.
    """

    def __init__(self, cache_size):
        self.cache = {}
        self.cache_size = cache_size

    def hash_summary(self, summary):
        return hashlib.sha256(summary.encode()).hexdigest()

    def add(self, summary, embedding):
        self.cache[self.hash_summary(summary)] = {
            "embedding": embedding,
            "used": datetime.datetime.now()
        }

        if len(self.cache) > self.cache_size:
            # Let's just lop off a quarter of it, arbitrarily
            for key in [ k for (k,v) in sorted(self.cache.items(), key=lambda item: item[1]["used"])][0:int(self.cache_size / 4)]:
                del self.cache[key]

    def get(self, summary):
        sha = self.hash_summary(summary)
        if sha in self.cache:
            resp = self.cache[sha]
            resp["used"] = datetime.datetime.now()
            return resp["embedding"]


embeddings_cache = EmbeddingsCache(getattr(settings, "SEMANTIC_SEARCH_EMBEDDINGS_CACHE_SIZE", 200))



import requests
from django.conf import settings
from sentence_transformers import SentenceTransformer


class LLM:
    def __init__(self):
        self.api_base = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_API_BASE_URL")
        self.api_key = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_API_KEY")
        self.model_name = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_MODEL")
        self.session = requests.Session()

    def get_embeddings(self, texts, *, prompt_name=None) -> list[list[float]]:
        headers = {"Authorization": f"Bearer {self.api_key}"}
        data = {"input": texts, "model": self.model_name}
        if prompt_name:
            data["type"] = prompt_name

        num_retries = 0
        while num_retries < 5:
            response = self.session.post(
                f"{self.api_base}/embeddings", json=data, headers=headers
            )
            if response.status_code == 200:
                break
            num_retries = num_retries + 1

        if num_retries == 5:
            raise Exception("get_embeddings call failed too many times")

        payload = response.json()
        results = payload.get("data", [])
        embeddings = [result.get("embedding") for result in results]

        return embeddings


class LocalLLM(LLM):
    def __init__(self):
        self.model_name = getattr(settings, "SEMANTIC_SEARCH_EMBEDDING_MODEL")
        self.model = SentenceTransformer(
            self.model_name,
            trust_remote_code=True,
            config_kwargs={"use_memory_efficient_attention": False},
        )

    def get_embeddings(self, texts, *, prompt_name=None):
        if isinstance(texts, str):
            texts = [texts]

        return self.model.encode(texts, prompt_name=prompt_name).tolist()


llm = LLM()
local_llm = LocalLLM()



# Generated by Django 4.1 on 2024-10-24 12:31

import django.db.models.deletion
from django.db import migrations, models
from pgvector.django import VectorExtension
from pgvector.django.indexes import HnswIndex
from pgvector.django.vector import VectorField


class Migration(migrations.Migration):
    initial = True

    dependencies = [
        ("torque", "0048_rename_search_cache_dirty_wikiconfig_cache_dirty"),
    ]

    operations = [
        VectorExtension(),
        migrations.CreateModel(
            name="SemanticSearchCacheDocument",
            fields=[
                (
                    "id",
                    models.AutoField(
                        auto_created=True,
                        primary_key=True,
                        serialize=False,
                        verbose_name="ID",
                    ),
                ),
                ("data", models.TextField()),
                (
                    "data_embedding",
                    VectorField(dimensions=768, null=True),
                ),
                (
                    "search_cache_document",
                    models.ForeignKey(
                        on_delete=django.db.models.deletion.CASCADE,
                        related_name="semantic_documents",
                        to="torque.searchcachedocument",
                    ),
                ),
            ],
        ),
        migrations.AddIndex(
            model_name="semanticsearchcachedocument",
            index=HnswIndex(
                ef_construction=64,
                fields=["data_embedding"],
                m=16,
                name="data_embedding_index",
                opclasses=["vector_cosine_ops"],
            ),
        ),
    ]



from django.conf import settings
from django.db import models
from pgvector.django import HnswIndex, VectorField
from torque.models import SearchCacheDocument


class SemanticSearchCacheDocument(models.Model):
    search_cache_document = models.ForeignKey(
        SearchCacheDocument,
        on_delete=models.CASCADE,
        related_name="semantic_documents",
    )
    data = models.TextField()
    data_embedding = VectorField(
        dimensions=getattr(settings, "SEMANTIC_SEARCH_NUM_DIMENSIONS", 768),
        null=True,
    )

    class Meta:
        indexes = [
            HnswIndex(
                name="data_embedding_index",
                fields=["data_embedding"],
                m=16,
                ef_construction=64,
                opclasses=["vector_cosine_ops"],
            ),
        ]



#!/usr/bin/env python

from distutils.core import setup

# We read version from here to make it easier to read for outside scripts
# (like release scripts)
main_ns = {}
with open("semantic_search/version.py") as ver_file:
    exec(ver_file.read(), main_ns)

with open("README.md", "r", encoding="utf-8") as readme:
    long_description = readme.read()

setup(
    name="django-torque-semantic-search",
    version=main_ns["__version__"],
    description="django app for torque semantic search",
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Open Tech Strategies, LLC",
    author_email="frankduncan@opentechstrategies.com",  # For now, this works
    url="https://code.librehq.com/ots/mediawiki/semantic-search",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: GNU Affero General Public License v3",
        "Operating System :: OS Independent",
    ],
    packages=[
        "semantic_search",
        "semantic_search.migrations",
    ],
    install_requires=[
        "django-torque",
        "pgvector",
        "orjson",
        "sentence-transformers",
        "transformers<4.51.0",
        "luqum",
    ],
    extras_require={
        "dev": ["pytest"],
    },
    package_dir={"": "."},
    python_requires=">=3.11",
)



from django.conf import settings
from luqum.parser import parser

from semantic_search.transformers import (
    FilterQueryTransformer,
    WithoutNegationsTransformer,
)


def filter_queries(qs, keep_negations=True):
    """
    Filters a list of queries to keep only phrases, e.g. "term", and
    negations, and some operators,
    e.g. `title:water water in india -"is a resource" -goats "test"`
    becomes `-"is a resource" -goats "test"`.
    Or just removes negations.
    """

    Transformer = (
        FilterQueryTransformer
        if keep_negations
        else WithoutNegationsTransformer
    )

    return [str(Transformer().visit(parser.parse(q))).strip() for q in qs]


def build_semantic_summary(document_dict, filtered_data):
    embedding_data = {}

    for filter in getattr(settings, "SEMANTIC_SEARCH_ADDITIONAL_FILTERS", []):
        embedding_data[filter.name()] = filter.document_value(document_dict)

    embedding_data.update(filtered_data)

    data_text = ""
    for name, value in embedding_data.items():
        name = name.replace("_", " ")
        if isinstance(value, list):
            for v in value:
                data_text += f"{name} is {v}. "
        elif value:
            data_text += f"{name} is {value}. "

    return data_text
